@@ -344,7 +344,7 @@ class XRImage:
             warnings.warn(
                 "include_scale_offset_tags is deprecated, please use "
                 "scale_offset_tags to indicate tag labels",
-                DeprecationWarning)
+                DeprecationWarning, stacklevel=2)
             scale_offset_tags = scale_offset_tags or ("scale", "offset")
 
         if tags is None:
@@ -659,7 +659,7 @@ class XRImage:
                     "Specified fill value will overlap with valid "
                     "data. To avoid this warning specify a fill_value "
                     "that is the minimum or maximum for the data type "
-                    "being saved to.")
+                    "being saved to.", stacklevel=3)
         return scale, offset
 
     def _scale_to_dtype(self, data, dtype, fill_value=None):
@@ -940,18 +940,28 @@ class XRImage:
         the normal [0,1] range of the channels.
 
         """
-        if isinstance(gamma, (list, tuple)):
-            gamma = self.xrify_tuples(gamma)
-        elif gamma is None or gamma == 1.0:
+        if _is_unity_or_none(gamma):
             return
 
+        inverse_gamma = self._get_inverse_gamma(gamma)
         logger.debug("Applying gamma %s", str(gamma))
         attrs = self.data.attrs
         self.data = self.data.clip(min=0)
-        self.data **= 1.0 / gamma
+        self.data **= inverse_gamma
         self.data.attrs = attrs
         self.data.attrs.setdefault('enhancement_history', []).append({'gamma': gamma})
 
+    def _get_inverse_gamma(self, gamma):
+        if np.issubdtype(self.data.dtype, np.floating):
+            dtype = self.data.dtype
+        else:
+            dtype = np.float32
+        if isinstance(gamma, (list, tuple)):
+            gamma = self.xrify_tuples(gamma).astype(dtype)
+        else:
+            gamma = np.array(gamma, dtype=dtype)
+        return 1.0 / gamma
+
     def stretch(self, stretch="crude", **kwargs):
         """Apply stretching to the current image.
 
@@ -1019,29 +1029,45 @@ class XRImage:
         """
         logger.debug("Perform a linear contrast stretch.")
 
+        left, right = self._get_left_and_right_quantiles_for_linear_stretch(cutoffs)
+
+        self.crude_stretch(left, right)
+
+    def _get_left_and_right_quantiles_for_linear_stretch(self, cutoffs):
         logger.debug("Calculate the histogram quantiles: ")
         logger.debug("Left and right quantiles: " +
                      str(cutoffs[0]) + " " + str(cutoffs[1]))
-
         cutoff_type = np.float64
         # numpy percentile (which quantile calls) returns 64-bit floats
         # unless the value is a higher order float
         if np.issubdtype(self.data.dtype, np.floating) and \
                 np.dtype(self.data.dtype).itemsize > 8:
             cutoff_type = self.data.dtype
-        left, right = dask.delayed(self._compute_quantile, nout=2)(self.data.data, self.data.dims, cutoffs)
-        left_data = da.from_delayed(left,
-                                    shape=(self.data.sizes['bands'],),
-                                    dtype=cutoff_type)
+
+        data = self.data
+        if 'A' in self.data.coords['bands'].values:
+            data = self.data.sel(bands=self.data.coords['bands'].values[:-1])
+
+        left_data, right_data = self._get_left_and_right_quantiles_without_alpha(data, cutoffs, cutoff_type)
+
+        if 'A' in self.data.coords['bands'].values:
+            left_data = np.hstack([left_data, np.array([0])])
+            right_data = np.hstack([right_data, np.array([1])])
         left = xr.DataArray(left_data, dims=('bands',),
                             coords={'bands': self.data['bands']})
-        right_data = da.from_delayed(right,
-                                     shape=(self.data.sizes['bands'],),
-                                     dtype=cutoff_type)
         right = xr.DataArray(right_data, dims=('bands',),
                              coords={'bands': self.data['bands']})
+        return left, right
 
-        self.crude_stretch(left, right)
+    def _get_left_and_right_quantiles_without_alpha(self, data, cutoffs, cutoff_type):
+        left, right = dask.delayed(self._compute_quantile, nout=2)(data.data, data.dims, cutoffs)
+        left_data = da.from_delayed(left,
+                                    shape=(data.sizes['bands'],),
+                                    dtype=cutoff_type)
+        right_data = da.from_delayed(right,
+                                     shape=(data.sizes['bands'],),
+                                     dtype=cutoff_type)
+        return left_data, right_data
 
     def crude_stretch(self, min_stretch=None, max_stretch=None):
         """Perform simple linear stretching.
@@ -1050,32 +1076,48 @@ class XRImage:
         normalizes to the [0,1] range.
 
         """
-        if min_stretch is None:
-            non_band_dims = tuple(x for x in self.data.dims if x != 'bands')
-            min_stretch = self.data.min(dim=non_band_dims)
-        if max_stretch is None:
-            non_band_dims = tuple(x for x in self.data.dims if x != 'bands')
-            max_stretch = self.data.max(dim=non_band_dims)
+        min_stretch = self._check_stretch_value(min_stretch, kind='min')
+        max_stretch = self._check_stretch_value(max_stretch, kind='max')
+        scale_factor = self._get_scale_factor(min_stretch, max_stretch)
 
-        if isinstance(min_stretch, (list, tuple)):
-            min_stretch = self.xrify_tuples(min_stretch)
-        if isinstance(max_stretch, (list, tuple)):
-            max_stretch = self.xrify_tuples(max_stretch)
-
-        delta = (max_stretch - min_stretch)
-        if isinstance(delta, xr.DataArray):
-            # fillna if delta is NaN
-            scale_factor = (1.0 / delta).fillna(0)
-        else:
-            scale_factor = 1.0 / delta
         attrs = self.data.attrs
         offset = -min_stretch * scale_factor
-        self.data *= scale_factor
-        self.data += offset
+        self.data = np.multiply(self.data, scale_factor, dtype=scale_factor.dtype) + offset
         self.data.attrs = attrs
         self.data.attrs.setdefault('enhancement_history', []).append({'scale': scale_factor,
                                                                       'offset': offset})
 
+    def _check_stretch_value(self, val, kind='min'):
+        if val is None:
+            non_band_dims = tuple(x for x in self.data.dims if x != 'bands')
+            val = getattr(self.data, kind)(dim=non_band_dims)
+
+        if isinstance(val, (list, tuple)):
+            val = self.xrify_tuples(val)
+
+        try:
+            val = val.astype(self.data.dtype)
+        except AttributeError:
+            val = self.data.dtype.type(val)
+
+        return val
+
+    def _get_scale_factor(self, min_stretch, max_stretch):
+        delta = (max_stretch - min_stretch)
+        dtype = self._infer_scale_factor_dtype()
+        if isinstance(delta, xr.DataArray):
+            # fillna if delta is NaN
+            scale_factor = (1.0 / delta).fillna(0).astype(dtype)
+        else:
+            scale_factor = np.array(1.0 / delta, dtype=dtype)
+
+        return scale_factor
+
+    def _infer_scale_factor_dtype(self):
+        if np.issubdtype(self.data.dtype, np.integer):
+            return np.float32
+        return self.data.dtype
+
     def stretch_hist_equalize(self, approximate=False):
         """Stretch the current image's colors through histogram equalization.
 
@@ -1284,6 +1326,7 @@ class XRImage:
         if self.mode not in ("L", "LA"):
             raise ValueError("Image should be grayscale to colorize")
 
+        colormap = self._adjust_colormap_dtype(colormap)
         l_data = self._get_masked_floating_luminance_data()
         alpha = self.data.sel(bands=['A']) if self.mode == "LA" else None
         new_data = colormap.colorize(l_data.data)
@@ -1309,6 +1352,12 @@ class XRImage:
             'colormap': colormap,
         })
 
+    def _adjust_colormap_dtype(self, colormap):
+        if np.issubdtype(self.data.dtype, np.floating) and colormap.colors.dtype != self.data.dtype:
+            colormap.colors = colormap.colors.astype(self.data.dtype)
+            colormap.values = colormap.values.astype(self.data.dtype)
+        return colormap
+
     def _get_masked_floating_luminance_data(self):
         l_data = self.data.sel(bands='L')
         # mask any integer fields with _FillValue
@@ -1468,3 +1517,11 @@ class XRImage:
         b = io.BytesIO()
         self.pil_image().save(b, format='png')
         return b.getvalue()
+
+
+def _is_unity_or_none(gamma):
+    if gamma is None or gamma == 1.0:
+        return True
+    if not hasattr(gamma, "__iter__"):
+        return False
+    return all(g == 1.0 for g in gamma) or all(g is None for g in gamma)
